__all__ = ["search"]


# standard library
from logging import getLogger
from typing import List, Sequence


# dependencies
from arxiv import Search
from dateparser import parse


# submodules
from .article import Article
from .consts import CATEGORIES, KEYWORDS, START_DATE, END_DATE


# constants
ARXIV_DATE_FORMAT = "%Y%m%d%H%M%S"


# logger
logger = getLogger(__name__)


# runtime functions
def search(
    categories: Sequence[str] = CATEGORIES,
    keywords: Sequence[str] = KEYWORDS,
    start_date: str = START_DATE,
    end_date: str = END_DATE,
) -> List[Article]:
    """Search for articles in arXiv.

    Args:
        categories: arXiv categories.
        keywords: Keywords of the search.
        start_date: Start date of the search.
        end_date: End date of the search.

    Returns:
        Articles found with given conditions.

    """
    start_date = format_date(start_date)
    end_date = format_date(end_date)

    query = f"submittedDate:[{start_date} TO {end_date}]"

    if categories:
        sub = " OR ".join(f"cat:{cat}" for cat in categories)
        query += f" AND ({sub})"

    if keywords:
        sub = " OR ".join(f'abs:"{kwd}"' for kwd in keywords)
        query += f" AND ({sub})"

    logger.debug(f"Searched articles by: {query!r}")
    results = list(Search(query).results())
    logger.debug(f"Number of articles found: {len(results)}")
    return list(map(Article.from_arxiv_result, results))


def format_date(date_like: str) -> str:
    """Parse and format a date-like string."""
    if (dt := parse(date_like)) is not None:
        return dt.strftime(ARXIV_DATE_FORMAT)

    raise ValueError(f"Could not parse {date_like!r}.")



__all__ = [
    "CATEGORIES",
    "KEYWORDS",
    "START_DATE",
    "END_DATE",
    "SOURCE_LANG",
    "TARGET_LANG",
    "DEEPL_MODE",
    "N_CONCURRENT",
    "TIMEOUT",
    "DeepLMode",
]


# standard library
from typing import Literal


# constants
CATEGORIES = ("astro-ph.*",)
KEYWORDS = ()
START_DATE = "3 days ago at midnight in UTC"
END_DATE = "2 days ago at midnight in UTC"
SOURCE_LANG = "en"
TARGET_LANG = "ja"
DEEPL_MODE = "auto"
N_CONCURRENT = 2
TIMEOUT = 30.0


# type hints
DeepLMode = Literal["auto", "api", "browser"]



__all__ = []


# standard library
from logging import DEBUG, basicConfig, getLogger
from typing import Sequence


# dependencies
from fire import Fire


# submodules
from .apps import slack
from .arxiv import search
from .consts import (
    CATEGORIES,
    KEYWORDS,
    START_DATE,
    END_DATE,
    SOURCE_LANG,
    TARGET_LANG,
    DEEPL_MODE,
    N_CONCURRENT,
    TIMEOUT,
    DeepLMode,
)
from .deepl import translate


# logger
logger = getLogger(__name__)


# runtime functions
def configure_logging(debug: bool = False) -> None:
    """Configure logging format and level for CLI."""
    if debug:
        getLogger("arxiv_post").setLevel(DEBUG)

    basicConfig(
        datefmt="%Y-%m-%d %H:%M:%S",
        format="[%(asctime)s %(name)s %(levelname)s]: %(message)s",
    )


def cmd_slack(
    categories: Sequence[str] = CATEGORIES,
    keywords: Sequence[str] = KEYWORDS,
    start_date: str = START_DATE,
    end_date: str = END_DATE,
    source_lang: str = SOURCE_LANG,
    target_lang: str = TARGET_LANG,
    deepl_mode: DeepLMode = DEEPL_MODE,
    deepl_api_key: str = "",
    n_concurrent: int = N_CONCURRENT,
    timeout: float = TIMEOUT,
    slack_webhook_url: str = "",
    dry_run: bool = False,
    debug: bool = False,
) -> None:
    """Translate and post articles to Slack.

    Args:
        categories: Comma-separated arXiv categories.
        keywords: Comma-separated keywords for a search.
        start_date: Start date for a search (inclusive).
        end_date: End date for a search (exclusive).
        source_lang: Language of original text in articles.
        target_lang: Language of translated text in articles.
        deepl_mode: Translation mode (auto, api, browser) of DeepL.
        deepl_api_key: Authentication Key for DeepL API (api-mode only).
        n_concurrent: Number of concurrent translation (browser-mode only).
        timeout: Timeout for translation in seconds (browser-mode only).
        slack_webhook_url: URL of Slack incoming webhook.
        dry_run: If True, articles are not posted to Slack.
        debug: If True, debug-level log messages are shown.

    Returns:
        This command does not return anything.

    """
    configure_logging(debug)

    if isinstance(categories, str):
        categories = categories.split(",")

    if isinstance(keywords, str):
        keywords = keywords.split(",")

    for name, value in locals().items():
        logger.debug(f"{name}: {value!r}")

    articles = search(
        categories=categories,
        keywords=keywords,
        start_date=start_date,
        end_date=end_date,
    )
    translated = translate(
        articles,
        source_lang=source_lang,
        target_lang=target_lang,
        deepl_mode=deepl_mode,
        deepl_api_key=deepl_api_key,
        n_concurrent=n_concurrent,
        timeout=timeout,
    )

    return slack.post(translated, slack_webhook_url, dry_run)


def main() -> None:
    """Entry point of command line interface."""
    Fire(dict(slack=cmd_slack))



__all__ = ["translate"]


# standard library
from asyncio import gather, sleep, run
from logging import getLogger
from textwrap import shorten
from typing import Iterable, List, Sequence, Protocol, TypeVar, cast


# dependencies
from deepl import AuthorizationException, Language, TextResult, Translator
from more_itertools import divide, flatten
from playwright.async_api import Page, async_playwright


# submodules
from .consts import (
    SOURCE_LANG,
    TARGET_LANG,
    DEEPL_MODE,
    N_CONCURRENT,
    TIMEOUT,
    DeepLMode,
)


# constants
DEEPL_INPUT = "textarea.lmt__source_textarea"
DEEPL_OUTPUT = "#target-dummydiv"
DEEPL_TRANSLATOR = "https://deepl.com/translator"


# logger
logger = getLogger(__name__)


# type hints
T = TypeVar("T")


class Translatable(Protocol):
    """Type hint for translatable objects."""

    def replace(self: T, original: str, translated: str, /) -> T:
        ...

    def __str__(self) -> str:
        ...


TL = TypeVar("TL", bound=Translatable)


# runtime functions
def translate(
    translatables: Sequence[TL],
    target_lang: str = TARGET_LANG,
    source_lang: str = SOURCE_LANG,
    deepl_mode: DeepLMode = DEEPL_MODE,
    deepl_api_key: str = "",
    n_concurrent: int = N_CONCURRENT,
    timeout: float = TIMEOUT,
) -> List[TL]:
    """Translate objects written in one language to another.

    Args:
        translatables: Translatable objects.
        target_lang: Language of the translated objects.
        source_lang: Language of the original objects.
        deepl_mode: Translation mode (auto, api, browser) of DeepL.
        deepl_api_key: Authentication Key for DeepL API (api-mode only).
        n_concurrent: Number of concurrent translation (browser-mode only).
        timeout: Timeout for translation in seconds (browser-mode only).

    Returns:
        Translated objects.

    """
    if not translatables:
        return list(translatables)

    target_lang = parse_language(target_lang)
    source_lang = parse_language(source_lang)

    if source_lang == target_lang:
        return list(translatables)

    if deepl_mode == "auto":
        deepl_mode = get_deepl_mode(translatables, deepl_api_key)
        logger.debug(f"Selected translation mode of DeepL: {deepl_mode!r}")

    if deepl_mode == "api":
        return translate_by_api(**locals())
    elif deepl_mode == "browser":
        return translate_by_browser(**locals())
    else:
        raise ValueError(f"{deepl_mode!r} is not supported.")


def translate_by_api(
    translatables: Sequence[TL],
    target_lang: str = TARGET_LANG,
    source_lang: str = SOURCE_LANG,
    deepl_api_key: str = "",
    **_,
) -> List[TL]:
    """Translate objects by the DeepL API."""
    deepl = Translator(deepl_api_key)

    sources = list(map(str, translatables))
    results = deepl.translate_text(
        sources,
        target_lang=target_lang,
        source_lang=source_lang,
    )

    results = cast(List[TextResult], results)
    zipped = zip(translatables, sources, results)
    return [tl.replace(src, res.text) for tl, src, res in zipped]


def translate_by_browser(
    translatables: Sequence[TL],
    target_lang: str = TARGET_LANG,
    source_lang: str = SOURCE_LANG,
    n_concurrent: int = N_CONCURRENT,
    timeout: float = TIMEOUT,
    **_,
) -> List[TL]:
    """Translate objects by the DeepL web translator."""
    n_concurrent = min(n_concurrent, len(translatables))
    url = f"{DEEPL_TRANSLATOR}#{source_lang}/{target_lang}/"

    async def main() -> List[TL]:
        async with async_playwright() as playwright:
            browser = await playwright.chromium.launch()
            context = await browser.new_context()
            context.set_default_timeout(1e3 * timeout)

            async def inner(translatables: Iterable[TL]) -> List[TL]:
                page = await context.new_page()

                try:
                    await page.goto(url)
                    return [
                        await _translate(translatable, page, timeout)
                        for translatable in translatables
                    ]
                finally:
                    await page.close()

            try:
                coros = map(inner, divide(n_concurrent, translatables))
                return list(flatten(await gather(*coros)))
            finally:
                await browser.close()

    return run(main())


async def _translate(translatable: TL, page: Page, timeout: float) -> TL:
    """Translate an object in a page of a browser."""
    if not (original := str(translatable)):
        return translatable

    await page.fill(DEEPL_INPUT, "")
    await page.fill(DEEPL_INPUT, original)

    for _ in range(int(timeout / 0.5)):
        await sleep(0.5)

        if (content := await page.text_content(DEEPL_OUTPUT)) is None:
            continue

        if not (content := content.strip()):
            continue

        try:
            return translatable.replace(original, content)
        except ValueError:
            break

    logger.warn(f"Failed to translate: {shorten(original, 50)!r}")
    return translatable


def parse_language(lang: str) -> str:
    """Parse and format a language string."""
    if (lang := lang.lower()) in vars(Language).values():
        return lang

    try:
        return getattr(Language, lang.upper())
    except AttributeError:
        raise ValueError(f"{lang!r} is not supported.")


def get_deepl_mode(translatables: Iterable[TL], deepl_api_key: str) -> DeepLMode:
    """Choose translation mode of DeepL."""
    if not deepl_api_key:
        return "browser"

    try:
        deepl = Translator(deepl_api_key)
        usage = deepl.get_usage().character
    except AuthorizationException:
        return "browser"

    if not (usage.count is None or usage.limit is None):
        rest = usage.limit - usage.count
        logger.debug(f"Number of character(s) left for translation: {rest}")

        if rest < sum(map(len, map(str, translatables))):
            return "browser"

    return "api"



__all__ = ["post"]


# standard library
from logging import getLogger
from typing import Any, Dict, Sequence


# dependencies
from requests import post as _post
from tomli import TOMLDecodeError, loads


# submodules
from ..article import Article


# constants
PAYLOAD_TOML = '''
text = """{header}"""

[[blocks]]
type = "header"

[blocks.text]
type = "plain_text"
text = """{header}"""

[[blocks]]
type = "section"

[blocks.text]
type = "mrkdwn"
text = """*Titie:* {title}"""

[[blocks]]
type = "section"

[blocks.text]
type = "mrkdwn"
text = """*Authors:* {authors}"""

[[blocks]]
type = "section"

[blocks.text]
type = "mrkdwn"
text = """*Summary:* {summary}"""

[[blocks]]
type = "actions"

[[blocks.elements]]
type = "button"
action_id = "view_arxiv"
url = """{arxiv_url}"""

[blocks.elements.text]
type = "plain_text"
text = "View arXiv"

[[blocks.elements]]
type = "button"
action_id = "view_pdf"
url = """{arxiv_pdf_url}"""

[blocks.elements.text]
type = "plain_text"
text = "View PDF"
'''


# logger
logger = getLogger(__name__)


# runtime functions
def post(articles: Sequence[Article], slack_webhook_url: str, dry_run: bool) -> None:
    """Post articles to Slack."""
    for article in articles:
        try:
            payload = to_payload(article)

            if not dry_run:
                _post(slack_webhook_url, json=payload)

            logger.debug(f"Posted an article: {article.arxiv_url}")
        except TOMLDecodeError:
            logger.warn(f"Failed to post an article: {article.arxiv_url}")


def to_payload(article: Article) -> Dict[str, Any]:
    """Convert an article object to a Slack payload."""
    if article.original is None:
        original = article
    else:
        original = article.original

    return loads(
        PAYLOAD_TOML.format(
            header=article.title,
            title=original.title,
            authors=", ".join(original.authors),
            summary=article.summary,
            arxiv_url=original.arxiv_url,
            arxiv_pdf_url=original.arxiv_pdf_url,
        )
    )



__all__ = ["Article"]


# standard library
from dataclasses import dataclass, field, replace
from typing import List, Optional


# dependencies
import re
from arxiv import Result
from pylatexenc.latex2text import LatexNodes2Text


# dataclasses
@dataclass
class Article:
    """Dataclass for arXiv articles."""

    title: str
    """Title of the article."""

    authors: List[str]
    """Author list of the article."""

    summary: str
    """Summary of the article."""

    arxiv_url: str
    """arXiv URL of the article."""

    original: Optional["Article"] = field(default=None, compare=False)
    """Original article before translation (if any)."""

    @property
    def arxiv_pdf_url(self) -> str:
        """arXiv PDF URL of the article."""
        return self.arxiv_url.replace("abs", "pdf")

    @classmethod
    def from_arxiv_result(cls, result: Result) -> "Article":
        """Create an article from an arXiv query result."""
        return Article(
            title=result.title,
            authors=[a.name for a in result.authors],
            summary=result.summary,
            arxiv_url=str(result),
        )

    def replace(self, original: str, translated: str) -> "Article":
        """Text replacement method for translation."""
        title, summary = translated.split("\n", 1)
        return replace(self, title=title, summary=summary, original=self)

    def __str__(self) -> str:
        """Text output method for translation."""
        return f"{self.title}\n{self.summary}"

    def __post_init__(self) -> None:
        """Remove TeX's control commands from texts."""
        self.title = detex(self.title)
        self.summary = detex(self.summary)


# runtime functions
def detex(text: str) -> str:
    """Remove TeX's control commands from a text."""
    text = re.sub(r"(\n+\s*|\n*\s+)", " ", text)
    return LatexNodes2Text(keep_comments=True).latex_to_text(text)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arxiv_post', 'arxiv_post.apps']

package_data = \
{'': ['*']}

install_requires = \
['arxiv>=1.4,<2.0',
 'dateparser>=1.1,<2.0',
 'deepl>=1.4,<2.0',
 'fire>=0.4,<0.5',
 'more-itertools>=8.13,<9.0',
 'playwright>=1.18,<2.0',
 'pylatexenc>=2.10,<3.0',
 'requests>=2.27,<3.0',
 'tomli>=2.0,<3.0']

entry_points = \
{'console_scripts': ['arxiv-post = arxiv_post.cli:main']}

setup_kwargs = {
    'name': 'arxiv-post',
    'version': '0.6.4',
    'description': 'Translate and post arXiv articles to Slack and various apps',
    'long_description': '# arxiv-post\n\n[![Release](https://img.shields.io/pypi/v/arxiv-post?label=Release&color=cornflowerblue&style=flat-square)](https://pypi.org/project/arxiv-post/)\n[![Python](https://img.shields.io/pypi/pyversions/arxiv-post?label=Python&color=cornflowerblue&style=flat-square)](https://pypi.org/project/arxiv-post/)\n[![Downloads](https://img.shields.io/pypi/dm/arxiv-post?label=Downloads&color=cornflowerblue&style=flat-square)](https://pepy.tech/project/arxiv-post)\n[![DOI](https://img.shields.io/badge/DOI-10.5281/zenodo.6127352-cornflowerblue?style=flat-square)](https://doi.org/10.5281/zenodo.6127352)\n[![Tests](https://img.shields.io/github/workflow/status/astropenguin/arxiv-post/Tests?label=Tests&style=flat-square)](https://github.com/astropenguin/arxiv-post/actions)\n\nTranslate and post arXiv articles to Slack and various apps\n\n## Installation\n\n```shell\n$ pip install arxiv-post\n$ playwright install chromium\n```\n\n## Usage\n\nCommand line interface `arxiv-post` is available after installation, with which you can translate and post arXiv articles to various apps.\nNote that only `slack` app is currently available.\nYou need to [create a custom Slack app to get an URL of incoming webhook](https://slack.com/help/articles/115005265063-Incoming-webhooks-for-Slack).\n\n```shell\n$ arxiv-post slack --keywords deshima \\\n                   --categories astro-ph.IM \\\n                   --target_lang ja \\\n                   --slack_webhook_url <Slack webhook URL>\n```\n\nThe posted article looks like this.\n\n![arxiv-post-slack.png](https://raw.githubusercontent.com/astropenguin/arxiv-post/master/docs/_static/arxiv-post-slack.png)\n\nFor detailed information, see the built-in help by the following command.\n\n```shell\n$ arxiv-post slack --help\n```\n\n## Example\n\nIt would be nice to regularly run the command by some automation tools such as GitHub Actions.\nHere is a live example where daily arXiv articles in [astro-ph.GA](https://arxiv.org/list/astro-ph.GA/new), [astro-ph.IM](https://arxiv.org/list/astro-ph.IM/new), and [astro-ph.HE](https://arxiv.org/list/astro-ph.HE/new) are posted to different channels of a Slack workspace.\n\n- [a-lab-nagoya/astro-ph-slack: Translate and post arXiv articles to Slack](https://github.com/a-lab-nagoya/astro-ph-slack)\n\n## References\n\n- [fkubota/Carrier-Owl: arxiv--> DeepL --> Slack](https://github.com/fkubota/Carrier-Owl): The arxiv-post package is highly inspired by their work.\n',
    'author': 'Akio Taniguchi',
    'author_email': 'taniguchi@a.phys.nagoya-u.ac.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/astropenguin/arxiv-post/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
